/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.util;

import java.util.*;
import java.util.stream.Collectors;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import com.google.common.base.Joiner;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityLiving;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.InventoryPlayer;
import net.minecraft.item.ItemStack;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.math.*;
import net.minecraft.world.EnumSkyBlock;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraftforge.fml.common.network.ByteBufUtils;
import net.minecraftforge.registries.IForgeRegistryEntry;

import io.netty.buffer.ByteBuf;
import net.dries007.tfc.Constants;
import net.dries007.tfc.api.registries.TFCRegistries;
import net.dries007.tfc.api.types.Plant;
import net.dries007.tfc.api.types.Rock;
import net.dries007.tfc.api.util.TFCConstants;
import net.dries007.tfc.objects.blocks.BlockPeat;
import net.dries007.tfc.objects.blocks.BlocksTFC;
import net.dries007.tfc.objects.blocks.plants.BlockShortGrassTFC;
import net.dries007.tfc.objects.blocks.stone.BlockRockVariant;
import net.dries007.tfc.objects.entity.EntitySeatOn;
import net.dries007.tfc.world.classic.ClimateTFC;
import net.dries007.tfc.world.classic.chunkdata.ChunkDataTFC;

public final class Helpers
{
    private static final Joiner JOINER_DOT = Joiner.on('.');

    public static void spreadGrass(World world, BlockPos pos, IBlockState us, Random rand)
    {
        if (world.func_175671_l(pos.func_177984_a()) < 4 && world.func_180495_p(pos.func_177984_a()).getLightOpacity(world, pos.func_177984_a()) > 2)
        {
            if (us.func_177230_c() instanceof BlockPeat)
            {
                world.func_175656_a(pos, BlocksTFC.PEAT.func_176223_P());
            }
            else if (us.func_177230_c() instanceof BlockRockVariant)
            {
                BlockRockVariant block = ((BlockRockVariant) us.func_177230_c());
                world.func_175656_a(pos, block.getVariant(block.getType().getNonGrassVersion()).func_176223_P());
            }
        }
        else
        {
            if (world.func_175671_l(pos.func_177984_a()) < 9) return;

            for (int i = 0; i < 4; ++i)
            {
                BlockPos target = pos.func_177982_a(rand.nextInt(3) - 1, rand.nextInt(5) - 3, rand.nextInt(3) - 1);
                if (world.func_189509_E(target) || !world.func_175667_e(target)) return;
                BlockPos up = target.func_177982_a(0, 1, 0);

                IBlockState current = world.func_180495_p(target);
                if (!BlocksTFC.isSoil(current) || BlocksTFC.isGrass(current)) continue;
                if (world.func_175671_l(up) < 4 || world.func_180495_p(up).getLightOpacity(world, up) > 3)
                    continue;

                if (current.func_177230_c() instanceof BlockPeat)
                {
                    world.func_175656_a(target, BlocksTFC.PEAT_GRASS.func_176223_P());
                }
                else if (current.func_177230_c() instanceof BlockRockVariant)
                {
                    Rock.Type spreader = Rock.Type.GRASS;
                    if ((us.func_177230_c() instanceof BlockRockVariant) && ((BlockRockVariant) us.func_177230_c()).getType() == Rock.Type.DRY_GRASS)
                        spreader = Rock.Type.DRY_GRASS;

                    BlockRockVariant block = ((BlockRockVariant) current.func_177230_c());
                    world.func_175656_a(target, block.getVariant(block.getType().getGrassVersion(spreader)).func_176223_P());
                }
            }

            for (Plant plant : TFCRegistries.PLANTS.getValuesCollection())
            {
                if (plant.getPlantType() == Plant.PlantType.SHORT_GRASS && rand.nextFloat() < 0.5f)
                {
                    float temp = ClimateTFC.getHeightAdjustedTemp(world, pos.func_177984_a());
                    BlockShortGrassTFC plantBlock = BlockShortGrassTFC.get(plant);

                    if (world.func_175623_d(pos.func_177984_a()) &&
                        plant.isValidLocation(temp, ChunkDataTFC.getRainfall(world, pos.func_177984_a()), Math.subtractExact(world.func_175642_b(EnumSkyBlock.SKY, pos.func_177984_a()), world.func_175657_ab())) &&
                        plant.isValidGrowthTemp(temp) &&
                        rand.nextDouble() < plantBlock.getGrowthRate(world, pos.func_177984_a()))
                    {
                        world.func_175656_a(pos.func_177984_a(), plantBlock.func_176223_P());
                    }
                }
            }
        }
    }

    /**
     * Makes an entity sit on a block
     *
     * @param world    the worldObj
     * @param pos      the BlockPos of the block to sit on
     * @param creature the entityLiving that will sit on this block
     * @param yOffset  the y offset of the top facing
     */
    public static void sitOnBlock(World world, BlockPos pos, EntityLiving creature, double yOffset)
    {
        if (!world.field_72995_K && !world.func_180495_p(pos).func_185904_a().func_76222_j())
        {
            EntitySeatOn seat = new EntitySeatOn(world, pos, yOffset);
            world.func_72838_d(seat);
            creature.func_184220_m(seat);
        }
    }

    /**
     * Returns the entity which is sitting on this BlockPos.
     *
     * @param world the WorldObj
     * @param pos   the BlockPos of this block
     * @return the entity which is sitting on this block, or null if none
     */
    @Nullable
    public static Entity getSittingEntity(World world, BlockPos pos)
    {
        if (!world.field_72995_K)
        {
            List<EntitySeatOn> seats = world.func_72872_a(EntitySeatOn.class, new AxisAlignedBB(pos).func_186662_g(1D));
            for (EntitySeatOn seat : seats)
            {
                if (seat.getPos().equals(pos))
                {
                    return seat.getSittingEntity();
                }
            }
        }
        return null;
    }

    /**
     * Copy from Item#rayTrace
     * Returns a RayTraceResult containing first found block in Players reach.
     *
     * @param worldIn    the world obj player stands in.
     * @param playerIn   the player obj
     * @param useLiquids do fluids counts as block?
     */
    @Nullable
    public static RayTraceResult rayTrace(World worldIn, EntityPlayer playerIn, boolean useLiquids)
    {
        Vec3d playerVec = new Vec3d(playerIn.field_70165_t, playerIn.field_70163_u + playerIn.func_70047_e(), playerIn.field_70161_v);
        float cosYaw = MathHelper.func_76134_b(-playerIn.field_70177_z * 0.017453292F - (float) Math.PI);
        float sinYaw = MathHelper.func_76126_a(-playerIn.field_70177_z * 0.017453292F - (float) Math.PI);
        float cosPitch = -MathHelper.func_76134_b(-playerIn.field_70125_A * 0.017453292F);
        float sinPitch = MathHelper.func_76126_a(-playerIn.field_70125_A * 0.017453292F);
        double reachDistance = playerIn.func_110148_a(EntityPlayer.REACH_DISTANCE).func_111126_e();
        Vec3d targetVec = playerVec.func_72441_c((sinYaw * cosPitch) * reachDistance, sinPitch * reachDistance, (cosYaw * cosPitch) * reachDistance);
        return worldIn.func_147447_a(playerVec, targetVec, useLiquids, !useLiquids, false);
    }

    /**
     * Copied from {@link net.minecraft.entity.Entity#rayTrace(double, float)} as it is client only
     *
     * @param blockReachDistance the reach distance
     * @param partialTicks       idk
     * @return the ray trace result
     */
    @Nullable
    public static RayTraceResult rayTrace(Entity entity, double blockReachDistance, float partialTicks)
    {
        Vec3d eyePosition = entity.func_174824_e(partialTicks);
        Vec3d lookVector = entity.func_70676_i(partialTicks);
        Vec3d rayTraceVector = eyePosition.func_72441_c(lookVector.field_72450_a * blockReachDistance, lookVector.field_72448_b * blockReachDistance, lookVector.field_72449_c * blockReachDistance);
        return entity.field_70170_p.func_147447_a(eyePosition, rayTraceVector, false, false, true);
    }

    public static boolean containsAnyOfCaseInsensitive(Collection<String> input, String... items)
    {
        Set<String> itemsSet = Arrays.stream(items).map(String::toLowerCase).collect(Collectors.toSet());
        return input.stream().map(String::toLowerCase).anyMatch(itemsSet::contains);
    }

    @SuppressWarnings("unchecked")
    public static <T extends TileEntity> T getTE(IBlockAccess world, BlockPos pos, Class<T> aClass)
    {
        TileEntity te = world.func_175625_s(pos);
        if (!aClass.isInstance(te)) return null;
        return (T) te;
    }

    public static String getEnumName(Enum<?> anEnum)
    {
        return JOINER_DOT.join(TFCConstants.MOD_ID, "enum", anEnum.getDeclaringClass().getSimpleName(), anEnum).toLowerCase();
    }

    public static String getTypeName(IForgeRegistryEntry<?> type)
    {
        //noinspection ConstantConditions
        return JOINER_DOT.join(TFCConstants.MOD_ID, "types", type.getRegistryType().getSimpleName(), type.getRegistryName().func_110623_a()).toLowerCase();
    }

    public static boolean playerHasItemMatchingOre(InventoryPlayer playerInv, String ore)
    {
        for (ItemStack stack : playerInv.field_70462_a)
        {
            if (!stack.func_190926_b() && OreDictionaryHelper.doesStackMatchOre(stack, ore))
            {
                return true;
            }
        }
        for (ItemStack stack : playerInv.field_70460_b)
        {
            if (!stack.func_190926_b() && OreDictionaryHelper.doesStackMatchOre(stack, ore))
            {
                return true;
            }
        }
        for (ItemStack stack : playerInv.field_184439_c)
        {
            if (!stack.func_190926_b() && OreDictionaryHelper.doesStackMatchOre(stack, ore))
            {
                return true;
            }
        }
        return false;
    }

    @Nonnull
    public static ItemStack consumeItem(ItemStack stack, int amount)
    {
        if (stack.func_190916_E() <= amount)
        {
            return ItemStack.field_190927_a;
        }
        stack.func_190918_g(amount);
        return stack;
    }

    @Nonnull
    public static ItemStack consumeItem(ItemStack stack, EntityPlayer player, int amount)
    {
        return player.func_184812_l_() ? stack : consumeItem(stack, amount);
    }

    public static void damageItem(ItemStack stack)
    {
        damageItem(stack, 1);
    }

    /**
     * Utility method for damaging an item that doesn't take an entity
     *
     * @param stack the stack to be damaged
     */
    public static void damageItem(ItemStack stack, int amount)
    {
        if (stack.func_96631_a(amount, Constants.RNG, null))
        {
            stack.func_190918_g(1);
            stack.func_77964_b(0);
        }
    }

    /**
     * Simple method to spawn items in the world at a precise location, rather than using InventoryHelper
     */
    public static void spawnItemStack(World world, BlockPos pos, ItemStack stack)
    {
        if (stack.func_190926_b())
            return;
        EntityItem entityitem = new EntityItem(world, pos.func_177958_n(), pos.func_177956_o(), pos.func_177952_p(), stack);
        world.func_72838_d(entityitem);
    }

    /**
     * Method for hanging blocks to check if they can hang. 11/10 description.
     * NOTE: where applicable, remember to still check if the blockstate allows for the specified direction!
     *
     * @param pos    position of the block that makes the check
     * @param facing the direction the block is facing. This is the direction the block should be pointing and the side it hangs ON, not the side it sticks WITH.
     *               e.g: a sign facing north also hangs on the north side of the support block
     * @return true if the side is solid, false otherwise.
     */
    public static boolean canHangAt(World worldIn, BlockPos pos, EnumFacing facing)
    {
        return worldIn.isSideSolid(pos.func_177972_a(facing.func_176734_d()), facing);
    }

    /**
     * Primarily for use in placing checks. Determines a solid side for the block to attach to.
     *
     * @param pos             position of the block/space to be checked.
     * @param possibleSides   a list/array of all sides the block can attach to.
     * @param preferredFacing this facing is checked first. It can be invalid or null.
     * @return Found facing or null is none is found. This is the direction the block should be pointing and the side it stick TO, not the side it sticks WITH.
     */
    public static EnumFacing getASolidFacing(World worldIn, BlockPos pos, @Nullable EnumFacing preferredFacing, EnumFacing... possibleSides)
    {
        return getASolidFacing(worldIn, pos, preferredFacing, Arrays.asList(possibleSides));
    }

    public static EnumFacing getASolidFacing(World worldIn, BlockPos pos, @Nullable EnumFacing preferredFacing, Collection<EnumFacing> possibleSides)
    {
        if (preferredFacing != null && possibleSides.contains(preferredFacing) && canHangAt(worldIn, pos, preferredFacing))
        {
            return preferredFacing;
        }
        for (EnumFacing side : possibleSides)
        {
            if (side != null && canHangAt(worldIn, pos, side))
            {
                return side;
            }
        }
        return null;
    }

    public static void writeResourceLocation(ByteBuf buf, @Nullable ResourceLocation loc)
    {
        buf.writeBoolean(loc != null);
        if (loc != null)
        {
            ByteBufUtils.writeUTF8String(buf, loc.toString());
        }
    }


    @Nullable
    public static ResourceLocation readResourceLocation(ByteBuf buf)
    {
        if (buf.readBoolean())
        {
            return new ResourceLocation(ByteBufUtils.readUTF8String(buf));
        }
        return null;
    }

    /**
     * This is meant to avoid Intellij's warnings about null fields that are injected to at runtime
     * Use this for things like @ObjectHolder, @CapabilityInject, etc.
     * AKA - The @Nullable is intentional. If it crashes your dev env, then fix your dev env, not this. :)
     *
     * @param <T> anything and everything
     * @return null, but not null
     */
    @Nonnull
    @SuppressWarnings("ConstantConditions")
    public static <T> T getNull()
    {
        return null;
    }
}
